#define IO_REGION(name, addr) (*(volatile struct name *)(addr))

/* offsets to add to 4kB I/O regions for atomic bitwise manipulation */
#define HW_XOR(_a, _v) *((volatile unsigned *)(((char *)&(_a)) + 0x1000)) = (_v)
#define HW_OR(_a, _v) *((volatile unsigned *)(((char *)&(_a)) + 0x2000)) = (_v)
#define HW_ANDNOT(_a, _v) *((volatile unsigned *)(((char *)&(_a)) + 0x3000)) = (_v)

struct clocks_rec {			/* pg197 */
	unsigned gpout0_ctrl, gpout0_div, gpout0_selected;
	unsigned gpout1_ctrl, gpout1_div, gpout1_selected;
	unsigned gpout2_ctrl, gpout2_div, gpout2_selected;
	unsigned gpout3_ctrl, gpout3_div, gpout3_selected;
	unsigned ref_ctrl, ref_div, ref_selected;
	unsigned sys_ctrl, sys_div, sys_selected;
	unsigned peri_ctrl, peri_div, peri_selected;
	unsigned usb_ctrl, usb_div, usb_selected;
	unsigned adc_ctrl, adc_div, adc_selected;
	unsigned rtc_ctrl, rtc_div, rtc_selected;
	unsigned sys_resus_ctrl, sys_resus_status;
	unsigned fc0_ref_khz, fc0_min_khz, fc0_max_khz;
	unsigned fc0_delay, fc0_interval, fc0_src, fc0_status, fc0_result;
	unsigned wake_en0, wake_en1;
	unsigned sleep_en0, sleep_en1;
	unsigned enabled0, enabled1;
	unsigned intr, inte, intf, ints;
};
#define clocks IO_REGION(clocks_rec, 0x40008000)

struct resets_rec {		/* pg179 */
	unsigned reset, wdsel, reset_done;
};
#define resets IO_REGION(resets_rec, 0x4000c000)

struct xosc_rec {		/* pg221 */
	unsigned ctrl, status, dormant, startup;
	unsigned reserved[3], count;
};
#define xosc IO_REGION(xosc_rec, 0x40024000)

struct rosc_rec {		/* pg225 */
	unsigned ctrl, freqa, freqb, dormant;
	unsigned div, phase, status, randombit;
	unsigned count;
};
#define rosc IO_REGION(rosc_rec, 0x40060000)

struct io_bank0_rec {		/* pg244 */
	struct { unsigned status, ctrl; } gpio[30];
	unsigned intr0, intr1, intr2, intr3;
	unsigned proc0_inte0, proc0_inte1, proc0_inte2, proc0_inte3;
	unsigned proc0_intf0, proc0_intf1, proc0_intf2, proc0_intf3;
	unsigned proc0_ints0, proc0_ints1, proc0_ints2, proc0_ints3;
	unsigned proc1_inte0, proc1_inte1, proc1_inte2, proc1_inte3;
	unsigned proc1_intf0, proc1_intf1, proc1_intf2, proc1_intf3;
	unsigned proc1_ints0, proc1_ints1, proc1_ints2, proc1_ints3;
	unsigned dormant_wake_inte0, dormant_wake_inte1;
	unsigned dormant_wake_inte2, dormant_wake_inte3;
	unsigned dormant_wake_intf0, dormant_wake_intf1;
	unsigned dormant_wake_intf2, dormant_wake_intf3;
	unsigned dormant_wake_ints0, dormant_wake_ints1;
	unsigned dormant_wake_ints2, dormant_wake_ints3;
};
#define io_bank0 IO_REGION(io_bank0_rec, 0x40014000)

struct psm_rec {		/* pg173 */
	unsigned frce_on, frce_off, wdsel, done;
};
#define psm IO_REGION(psm_rec, 0x40010000)

struct wdt_rec {		/* pg548 */
	unsigned ctrl, load, reason, scratch[8], tick;
};
#define wdt IO_REGION(wdt_rec, 0x40058000)

struct systick_rec {		/* ARMv6 pg238 */
	unsigned csr, rvr, cvr, calib;
};
#define syst IO_REGION(systick_rec, 0xe000e010)

struct nvic_rec {		/* ARMv6 pg245 */
	unsigned iser, __res1[31];
	unsigned icer, __res2[31];
	unsigned ispr, __res3[31];
	unsigned icpr, __res4[31];
	unsigned ipr[8];
};
#define nvic IO_REGION(nvic_rec, 0xe000e100)

struct scb_rec {
	unsigned cpuid, icsr;
	void *vtor;
	unsigned aircr;
	unsigned scr, ccr, __res1, shpr2;
	unsigned shpr3, shcsr, __res2[2];
	unsigned dfsr;
};
#define scb IO_REGION(scb_rec, 0xe000ed00)

typedef void (*vfunc)(void);
struct vtor_rec {		/* ARMv6 pg192 */
	void *sp_main;		/* %sp at reset */
	vfunc reset, nmi, hardfault;
	unsigned __res1[7];
	vfunc svcall;
	unsigned __res2[2];
	vfunc pendsv, systick;
	vfunc ext[32];		/* from NVIC */
};
static struct vtor_rec __attribute__((aligned(128))) vtor = { 0 };

__asm(
".section .stack		/* create symbol for stack */\n"
".section .start\n"
".global start\n"
".type start,function\n"
"start:\n"
"	ldr r0,=.stack+4096	/* stack grows towards smaller addrs */\n"
"	mov sp,r0\n"
"	bl main\n"
"	bl reboot		/* just in case main() returns */\n"
"	.ltorg\n"
".text\n"
);


static void
xosc_init(void)
{
	xosc.ctrl = 0x00fabaa0;		/* ENABLE and 1_15MHZ */

	/* default XOSC_STARTUP is 4.2ms at 12MHz...datasheet suggests 1ms is
	 * enough. no need to change it. */

	while (!(xosc.status & 0x80000000));	/* wait for XOSC STABLE */

	clocks.ref_ctrl = 2;		/* clk_ref from XOSC */
	/* clk_sys etc will follow clk_ref */

	rosc.ctrl = 0x00d1e000;		/* disable ROSC */

/* XXX - to enable USB, set CLK_USB_CTRL to 0x00000840 ENABLE|clksrc_pll_usb,
after bringing up that PLL at 4x */
}

static void
wdt_init(void)
{
	wdt.load = 0x00ffffff;
	wdt.tick = 0x20c;	/* enable, cycles=12, 1us off 12MHz XOSC?? */

	/* 0x0001fff8 (reset everything but ROSC/XOSC/CLOCKS) works too but
	 * let's go for the heavier reboot */
	psm.wdsel = 0x0001ffff;			/* WDT resets everything */
	HW_OR(wdt.ctrl, 0x40000000);		/* enable */
}

static void
vtor_init(void)
{
	scb.vtor = &vtor;
}

/* reboot by enabling the WDT with a 50us countdown and then infinite looping
 * until the WDT hits 0 */
void
reboot(void)
{
	psm.wdsel = 0x0001ffff;	/* WDT resets everything */
	wdt.tick = 0x20c;	/* enable, cycles=12, 1us off 12MHz XOSC?? */
	wdt.load = 100;
	while (1) {
		__asm("");
	}
}

/* NB - rvr/cvr are only 24-bit so this is capped at 1398ms at 12MHz */
static void
delay(unsigned ms)
{
	syst.rvr = (12000000/1000)*ms;
	syst.cvr = 0;			/* reset systick */
	syst.csr = 0x00000005;		/* clksource=clk_sys, no interrupt, yes
					 * enabled */
	while (!(syst.csr & 0x00010000));	/* wait for COUNTFLAG */
}

/* register the callback to be called by the systick timer every so-many-ms (up
 * to 1398ms at 12MHz) */
static void
systick_register(unsigned ms, void (*cb)(void))
{
	vtor.systick = cb;

	syst.rvr = (12000000/1000)*ms;
	syst.cvr = 0;			/* reset systick */
	syst.csr = 0x00000007;		/* clksource=clk_sys, yes interrupt,
					 * yes enabled */
}

/* enables the LED drive, then led_toggle() to turn it on or off */
static void
led_init(void)
{
	HW_ANDNOT(resets.reset, 1<<5);	/* enable IO_BANK0 */
	/* OEOVER=3 enable output, OUTOVER=1 output low, FUNCSEL=31 (NULL) */
	io_bank0.gpio[25].ctrl = 0x0000321f;
}

static void
led_toggle(void)
{
	HW_XOR(io_bank0.gpio[25].ctrl,  0x00000100);
}


void
main(void)
{
	xosc_init();
	wdt_init();
	vtor_init();

	led_init();
	systick_register(333, led_toggle);
	while (1) {
		__asm("wfi");
	}
}
